@@ -5,9 +5,10 @@ Contains helpers for plotting.
 """
 from __future__ import absolute_import, annotations, division, print_function
 import datetime
+import logging
 import os
 from pathlib import Path
-from typing import Any, Tuple, Optional
+from typing import Any, Optional, Tuple
 import warnings
 
 import matplotlib.pyplot as plt
@@ -16,7 +17,9 @@ import numpy as np
 import pandas as pd
 import seaborn as sns
 import xarray as xr
-import logging
+
+# from l2hmc.experiment.pytorch.experiment import Experiment as ptExperiment
+# from l2hmc.experiment.tensorflow.experiment import Experiment as tfExperiment
 
 warnings.filterwarnings('ignore')
 
@@ -26,7 +29,7 @@ xplt = xr.plot  # type: ignore
 
 LW = plt.rcParams.get('axes.linewidth', 1.75)
 
-colors = {
+COLORS = {
     'blue':     '#007DFF',
     'red':      '#FF5252',
     'green':    '#63FF5B',
@@ -43,16 +46,19 @@ plt.rcParams.update({
     'image.cmap': 'viridis',
     'savefig.transparent': True,
     'text.color': '#666666',
-    'axes.labelcolor': '#666666',
     'xtick.color': '#66666604',
     'ytick.color': '#66666604',
     'ytick.labelcolor': '#666666',
     'xtick.labelcolor': '#666666',
     'axes.edgecolor': '#66666600',
-    'grid.color': (0.4, 0.4, 0.4, 0.4),  # #66666602
+    'axes.labelcolor': '#666666',
+    # 'axes.labelcolor': (189, 189, 189, 1.0),
+    'grid.color': (0.434, 0.434, 0.434, 0.2),  # #66666602
     'axes.facecolor': (1.0, 1.0, 1.0, 0.0),
     'figure.facecolor': (1.0, 1.0, 1.0, 0.0),
 })
+
+plt.rcParams['axes.labelcolor'] = '#bdbdbd'
 # sns.set_palette(list(colors.values()))
 # sns.set_context('notebook', font_scale=0.8)
 # plt.rcParams.update({
@@ -66,6 +72,27 @@ plt.rcParams.update({
 # })
 
 
+def set_plot_style():
+    plt.style.use('default')
+    plt.rcParams.update({
+        'image.cmap': 'viridis',
+        'savefig.transparent': True,
+        'text.color': '#666666',
+        'xtick.color': '#66666604',
+        'ytick.color': '#66666604',
+        'ytick.labelcolor': '#666666',
+        'xtick.labelcolor': '#666666',
+        'axes.edgecolor': '#66666600',
+        'axes.labelcolor': '#666666',
+        # 'axes.labelcolor': (189, 189, 189, 1.0),
+        'grid.color': (0.434, 0.434, 0.434, 0.2),  # #66666602
+        'axes.facecolor': (1.0, 1.0, 1.0, 0.0),
+        'figure.facecolor': (1.0, 1.0, 1.0, 0.0),
+    })
+
+    plt.rcParams['axes.labelcolor'] = '#bdbdbd'
+
+
 def get_timestamp(fstr=None):
     """Get formatted timestamp."""
     now = datetime.datetime.now()
@@ -77,14 +104,76 @@ def get_timestamp(fstr=None):
 FigAxes = Tuple[plt.Figure, plt.Axes]
 
 
+def save_figure(fig: plt.Figure, fname: str, outdir: os.PathLike):
+    pngdir = Path(outdir).joinpath('pngs')
+    pngdir.mkdir(exist_ok=True, parents=True)
+    pngfile = pngdir.joinpath(f'{fname}.png')
+    svgfile = Path(outdir).joinpath(f'{fname}.svg')
+    fig.savefig(pngfile, dpi=400, bbox_inches='tight')
+    fig.savefig(svgfile, dpi=400, bbox_inches='tight')
+
+
 def savefig(fig: plt.Figure, outfile: os.PathLike):
     fout = Path(outfile)
     parent = fout.parent
     parent.mkdir(exist_ok=True, parents=True)
-    print(f'Saving figure to: {fout.as_posix()}')
+    log.info(f'Saving figure to: {fout.as_posix()}')
     fig.savefig(fout.as_posix(), dpi=400, bbox_inches='tight')
 
 
+def measure_improvement(
+        experiment: Any,
+        title: Optional[str] = None,
+) -> None:
+    ehist = experiment.trainer.histories.get('eval', None)
+    hhist = experiment.trainer.histories.get('hmc', None)
+    if ehist is not None and hhist is not None:
+        edset = ehist.get_dataset()
+        hdset = hhist.get_dataset()
+        dQint_eval = edset.dQint.mean('chain')[1:]
+        dQint_hmc = hdset.dQint.mean('chain')[1:]
+        fig, ax = plt.subplots()
+        _ = ax.plot(
+            dQint_eval,
+            label='Trained',
+            lw=2.,
+            color=COLORS['blue'],
+        )
+        _ = ax.plot(
+            dQint_hmc,
+            label='HMC',
+            ls=':',
+            lw=1.5,
+            color=COLORS['blue'],
+        )
+        _ = ax.grid(True, alpha=0.2)
+        xticks = ax.get_xticks()
+        # xticklabels = ax.get_xticklabels()
+        _ = ax.set_xticklabels([
+            f'{experiment.config.steps.log * int(i)}' for i in xticks
+        ])
+        _ = ax.set_xlabel('MD Step')
+        _ = ax.set_ylabel('dQint')
+        _ = ax.legend(
+            loc='best',
+            framealpha=0.1,
+            ncol=2,
+            labelcolor='#FFF',
+            shadow=True
+        )
+        if title is not None:
+            _ = ax.set_title(title)
+
+        outdir = experiment._outdir
+        improvement = np.mean(dQint_eval.values / dQint_hmc.values)
+        txtfile = Path(outdir).joinpath('model_improvement.txt').as_posix()
+        log.warning(f'Writing model improvement to: {txtfile}')
+        with open(txtfile, 'w') as f:
+            f.write(f'{improvement:.8f}')
+
+        save_figure(fig, fname='model_improvement', outdir=outdir)
+
+
 def plot_scalar(
         y: np.ndarray,
         x: Optional[np.ndarray] = None,
@@ -171,6 +260,7 @@ def plot_leapfrogs(
         xlabel: Optional[str] = None,
         ylabel: Optional[str] = None,
         outfile: Optional[os.PathLike] = None,
+        line_labels: Optional[bool] = False,
 ) -> FigAxes:
     assert len(y.shape) == 3
 
@@ -190,7 +280,8 @@ def plot_leapfrogs(
     for lf in range(nlf):
         _ = ax.plot(x, yavg[:, lf], color=colors[lf], label=f'{lf}')
 
-    _ = matplotx.line_labels(font_kwargs={'size': 'small'})
+    if line_labels:
+        _ = matplotx.line_labels(font_kwargs={'size': 'small'})
 
     if xlabel is not None:
         _ = ax.set_xlabel(xlabel)
@@ -237,11 +328,13 @@ def plot_combined(
     sns.kdeplot(y=val.values.flatten(), ax=ax2, color=color, shade=True)
     axes = (ax1, ax2)
     ax0 = subfigs[0].subplots(1, 1)
-    val = val.dropna('chain')
+    if 'chain' in val.dims:
+        val = val.dropna('chain')
+        _ = xplt.imshow(val, 'draw', 'chain', ax=ax0,
+                        robust=True, add_colorbar=True)
     # _ = xplt.pcolormesh(val, 'draw', 'chain', ax=ax0,
     #                     robust=True, add_colorbar=True)
-    _ = xplt.imshow(val, 'draw', 'chain', ax=ax0,
-                    robust=True, add_colorbar=True)
+    # sns.despine(ax0)
     nchains = min(num_chains, len(val.coords['chain']))
     label = f'{key}_avg'
     # label = r'$\langle$' + f'{key} ' + r'$\rangle$'
@@ -258,6 +351,7 @@ def plot_combined(
 
     ax2.set_xticks([])
     ax2.set_xticklabels([])
+    # sns.despine(ax=ax0, top=True, right=True, left=True, bottom=True)
     sns.despine(ax=ax1, top=True, right=True, left=True, bottom=True)
     sns.despine(ax=ax2, top=True, right=True, left=True, bottom=True)
     ax2.set_xlabel('')
@@ -274,13 +368,15 @@ def plot_dataArray(
         therm_frac: Optional[float] = 0.,
         num_chains: Optional[int] = 10,
         title: Optional[str] = None,
-        outdir: Optional[str] = None,
+        outdir: Optional[str | Path] = None,
         subplots_kwargs: Optional[dict[str, Any]] = None,
         plot_kwargs: Optional[dict[str, Any]] = None,
-        line_labels: Optional[bool] = True,
+        line_labels: Optional[bool] = False,
 ) -> tuple:
     plot_kwargs = {} if plot_kwargs is None else plot_kwargs
     subplots_kwargs = {} if subplots_kwargs is None else subplots_kwargs
+    set_plot_style()
+    plt.rcParams['axes.labelcolor'] = '#bdbdbd'
     figsize = subplots_kwargs.get('figsize', set_size())
     subplots_kwargs.update({'figsize': figsize})
     subfigs = None
@@ -348,19 +444,29 @@ def plot_dataArray(
         fig.suptitle(title)
 
     if outdir is not None:
-        plt.savefig(Path(outdir).joinpath(f'{key}.svg'),
-                    dpi=400, bbox_inches='tight')
+        outfile = Path(outdir).joinpath(f'{key}.svg')
+        if outfile.is_file():
+            tstamp = get_timestamp('%Y-%m-%d-%H%M%S')
+            pngdir = Path(outdir).joinpath('pngs')
+            pngdir.mkdir(exist_ok=True, parents=True)
+            pngfile = pngdir.joinpath(f'{key}-{tstamp}.png')
+            svgfile = Path(outdir).joinpath(f'{key}-{tstamp}.svg')
+            plt.savefig(pngfile, dpi=400, bbox_inches='tight')
+            plt.savefig(svgfile, dpi=400, bbox_inches='tight')
+            # plt.savefig(Path(outdir).joinpath(f'{key}.svg'),
+            #             dpi=400, bbox_inches='tight')
 
     return (fig, subfigs, axes)
 
 
 def plot_array(
-    val: list | np.ndarray,
-    key: Optional[str] = None,
-    xlabel: Optional[str] = None,
-    title: Optional[str] = None,
-    num_chains: Optional[int] = 10,
-    **kwargs,
+        val: list | np.ndarray,
+        key: Optional[str] = None,
+        xlabel: Optional[str] = None,
+        title: Optional[str] = None,
+        num_chains: Optional[int] = 10,
+        outdir: Optional[str | Path] = None,
+        **kwargs,
 ) -> tuple[plt.Figure, plt.Axes]:
     fig, ax = plt.subplots(constrained_layout=True)
     arr = np.array(val)
@@ -397,7 +503,7 @@ def plot_array(
 
     elif len(arr.shape) == 1:
         y = arr
-        x = np.arange(len(y))
+        x = np.arange(y.shape[0])
         _ = ax.plot(x, y, label=key, **kwargs)
 
     else:
@@ -411,6 +517,17 @@ def plot_array(
 
     _ = ax.legend(loc='best')
 
+    if outdir is not None:
+        outfile = Path(outdir).joinpath(f'{key}.svg')
+        if outfile.is_file():
+            tstamp = get_timestamp('%Y-%m-%d-%H%M%S')
+            pngdir = Path(outdir).joinpath('pngs')
+            pngdir.mkdir(exist_ok=True, parents=True)
+            pngfile = pngdir.joinpath(f'{key}-{tstamp}.png')
+            svgfile = Path(outdir).joinpath(f'{key}-{tstamp}.svg')
+            plt.savefig(pngfile, dpi=400, bbox_inches='tight')
+            plt.savefig(svgfile, dpi=400, bbox_inches='tight')
+
     return fig, ax
 
 
@@ -454,6 +571,7 @@ def plot_metric(
         subplots_kwargs: Optional[dict[str, Any]] = None,
         plot_kwargs: Optional[dict[str, Any]] = None,
         ext: Optional[str] = 'png',
+        line_labels: Optional[bool] = False,
 ) -> tuple:
     plot_kwargs = {} if plot_kwargs is None else plot_kwargs
     subplots_kwargs = {} if subplots_kwargs is None else subplots_kwargs
@@ -513,6 +631,9 @@ def plot_metric(
             cmap = plt.get_cmap('viridis', lut=arr.shape[1])
             _ = plot_kwargs.pop('color', None)
             for idx in range(arr.shape[1]):
+                label = plot_kwargs.pop('label', None)
+                if label is not None:
+                    label = f'{label}-{idx}'
                 y = arr[:, idx]
                 color = cmap(idx / y.shape[1])
                 if len(y.shape) == 2:
@@ -523,10 +644,11 @@ def plot_metric(
                                     lw=LW/4., alpha=0.7, **plot_kwargs)
 
                     ax.plot(steps, y.mean(-1), color=color,
-                            label=f'{idx}', **plot_kwargs)
+                            label=label, **plot_kwargs)
                 else:
+
                     ax.plot(steps, y, color=color,
-                            label=f'{idx}', **plot_kwargs)
+                            label=label, **plot_kwargs)
             axes = ax
         else:
             raise ValueError('Unexpected shape encountered')
@@ -539,7 +661,9 @@ def plot_metric(
             # where arr[:, idx].shape = [ndraws, 1]
             ax.plot(steps, arr[:, idx], alpha=0.5, lw=lw/2., **plot_kwargs)
 
-    matplotx.line_labels(font_kwargs={'size': 'small'})
+    if line_labels:
+        matplotx.line_labels(font_kwargs={'size': 'small'})
+
     ax.set_xlabel('draw')
     if title is not None:
         fig.suptitle(title)
@@ -640,85 +764,97 @@ def make_ridgeplots(
         outdir: Optional[os.PathLike] = None,
         drop_zeros: Optional[bool] = False,
         drop_nans: Optional[bool] = True,
-        cmap: Optional[str] = 'viridis_r',
-        # default_style: dict = None,
+        cmap: Optional[str] = 'rainbow',
 ):
     """Make ridgeplots."""
     data = {}
     # with sns.axes_style('white', rc={'axes.facecolor': (0, 0, 0, 0)}):
-    sns.set(style='white', palette='bright', context='paper')
-    plt.rcParams['axes.facecolor'] = (0, 0, 0, 0.0)
-    plt.rcParams['figure.facecolor'] = (0, 0, 0, 0.0)
-    for key, val in dataset.data_vars.items():
-        if 'leapfrog' in val.coords.dims:
-            lf_data = {
-                key: [],
-                'lf': [],
-            }
-            for lf in val.leapfrog.values:
-                # val.shape = (chain, leapfrog, draw)
-                # x.shape = (chain, draw);  selects data for a single lf
-                x = val[{'leapfrog': lf}].values
-                # if num_chains is not None, keep `num_chains` for plotting
-                if num_chains is not None:
-                    x = x[:num_chains, :]
+    # sns.set(style='white', palette='bright', context='paper')
+    # with sns.set_style(style='white'):
+    outdir = Path(os.getcwd()) if outdir is None else Path(outdir)
+    outdir = outdir.joinpath('ridgeplots')
+    with sns.plotting_context(
+            context='paper',
+    ):
+        sns.set(
+            style='white',
+            palette='bright',
+        )
+        plt.rcParams['axes.facecolor'] = (0, 0, 0, 0.0)
+        plt.rcParams['figure.facecolor'] = (0, 0, 0, 0.0)
+        for key, val in dataset.data_vars.items():
+            if 'leapfrog' in val.coords.dims:
+                lf_data = {
+                    key: [],
+                    'lf': [],
+                }
+                for lf in val.leapfrog.values:
+                    # val.shape = (chain, leapfrog, draw)
+                    # x.shape = (chain, draw);  selects data for a single lf
+                    x = val[{'leapfrog': lf}].values
+                    # if num_chains is not None, keep `num_chains` for plotting
+                    if num_chains is not None:
+                        x = x[:num_chains, :]
 
-                x = x.flatten()
-                if drop_zeros:
-                    x = x[x != 0]
-                #  x = val[{'leapfrog': lf}].values.flatten()
-                if drop_nans:
-                    x = x[np.isfinite(x)]
+                    x = x.flatten()
+                    if drop_zeros:
+                        x = x[x != 0]
+                    #  x = val[{'leapfrog': lf}].values.flatten()
+                    if drop_nans:
+                        x = x[np.isfinite(x)]
 
-                lf_arr = np.array(len(x) * [f'{lf}'])
-                lf_data[key].extend(x)
-                lf_data['lf'].extend(lf_arr)
+                    lf_arr = np.array(len(x) * [f'{lf}'])
+                    lf_data[key].extend(x)
+                    lf_data['lf'].extend(lf_arr)
 
-            lfdf = pd.DataFrame(lf_data)
-            data[key] = lfdf
+                lfdf = pd.DataFrame(lf_data)
+                data[key] = lfdf
 
-            # Initialize the FacetGrid object
-            ncolors = len(val.leapfrog.values)
-            pal = sns.color_palette(cmap, n_colors=ncolors)
-            g = sns.FacetGrid(lfdf, row='lf', hue='lf',
-                              aspect=15, height=0.25, palette=pal)
+                # Initialize the FacetGrid object
+                ncolors = len(val.leapfrog.values)
+                pal = sns.color_palette(cmap, n_colors=ncolors)
+                g = sns.FacetGrid(
+                    lfdf,
+                    row='lf', hue='lf',
+                    aspect=15, height=0.25, palette=pal  # type:ignore
+                )
 
-            # Draw the densities in a few steps
-            _ = g.map(sns.kdeplot, key, cut=1,
-                      shade=True, alpha=0.7, linewidth=1.25)
-            _ = g.map(plt.axhline, y=0, lw=1.5, alpha=0.7, clip_on=False)
+                # Draw the densities in a few steps
+                _ = g.map(sns.kdeplot, key, cut=1,
+                          shade=True, alpha=0.7, linewidth=1.25)
+                _ = g.map(plt.axhline, y=0, lw=1.5, alpha=0.9, clip_on=False)
 
-            # Define and use a simple function to
-            # label the plot in axes coords:
-            def label(x, color, label):  # type:ignore #noqa
-                ax = plt.gca()
-                ax.set_ylabel('')
-                ax.set_yticks([])
-                ax.set_yticklabels([])
-                ax.text(0, 0.10, label, fontweight='bold', color=color,
-                        ha='left', va='center', transform=ax.transAxes,
-                        fontsize='small')
+                # Define and use a simple function to
+                # label the plot in axes coords:
+                def label(_, color, label):  # type:ignore #noqa
+                    ax = plt.gca()
+                    ax.set_ylabel('')
+                    ax.set_yticks([])
+                    ax.set_yticklabels([])
+                    ax.text(0, 0.10, label, fontweight='bold', color=color,
+                            ha='left', va='center', transform=ax.transAxes)
 
-            _ = g.map(label, key)
-            # Set the subplots to overlap
-            _ = g.fig.subplots_adjust(hspace=-0.75)
-            # Remove the axes details that don't play well with overlap
-            _ = g.set_titles('')
-            _ = g.set(yticks=[])
-            _ = g.set(yticklabels=[])
-            _ = g.despine(bottom=True, left=True)
-            if outdir is not None:
-                outdir = Path(outdir)
-                pngdir = outdir.joinpath('pngs')
-                fsvg = Path(outdir).joinpath(f'{key}_ridgeplot.svg')
-                fpng = Path(pngdir).joinpath(f'{key}_ridgeplot.png')
+                _ = g.map(label, key)
+                # Set the subplots to overlap
+                _ = g.fig.subplots_adjust(hspace=-0.75)
+                # Remove the axes details that don't play well with overlap
+                _ = g.set_titles('')
+                _ = g.set(yticks=[])
+                _ = g.set(yticklabels=[])
+                _ = g.set(xlabel=f'{key}')
+                _ = g.despine(bottom=True, left=True)
+                if outdir is not None:
+                    outdir = Path(outdir)
+                    pngdir = outdir.joinpath('pngs')
+                    fsvg = Path(outdir).joinpath(f'{key}_ridgeplot.svg')
+                    fpng = Path(pngdir).joinpath(f'{key}_ridgeplot.png')
 
-                outdir.mkdir(exist_ok=True, parents=True)
-                pngdir.mkdir(exist_ok=True, parents=True)
+                    outdir.mkdir(exist_ok=True, parents=True)
+                    pngdir.mkdir(exist_ok=True, parents=True)
 
-                log.info(f'Saving figure to: {fsvg.as_posix()}')
-                plt.savefig(fsvg.as_posix(), dpi=500, bbox_inches='tight')
-                plt.savefig(fpng.as_posix(), dpi=500, bbox_inches='tight')
+                    log.warning(f'Saving figure to: {fsvg.as_posix()}')
+                    plt.savefig(fsvg.as_posix(), dpi=500, bbox_inches='tight')
+                    plt.savefig(fpng.as_posix(), dpi=500, bbox_inches='tight')
 
         # plt.close('all')
 
